/*
 * This software is governed by the CeCILL-B license under French law and
 * abiding by the rules of distribution of free software. You can use, modify
 * and/ or redistribute the software under the terms of the CeCILL-B license as
 * circulated by CEA, CNRS and INRIA at the following URL "http://www.cecill.info".
 * 
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability.
 * 
 * In this respect, the user's attention is drawn to the risks associated with
 * loading, using, modifying and/or developing or reproducing the software by
 * the user in light of its specific status of free software, that may mean that
 * it is complicated to manipulate, and that also therefore means that it is
 * reserved for developers and experienced professionals having in-depth
 * computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling
 * the security of their systems and/or data to be ensured and, more generally,
 * to use and operate it in the same conditions as regards security.
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */

package fr.amap.commons.util.swing;

import fr.amap.commons.util.locales.AvailablesLocales;
import fr.amap.commons.util.properties.ProjectProperties;
import java.awt.Window;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import java.util.function.Consumer;
import javax.swing.ImageIcon;

/**
 * Base class of each view class.
 */
public abstract class DialogView extends org.jdesktop.swingx.JXDialog {
    
    /** The current locale bundle */
    protected java.util.ResourceBundle bundle;
    
    /** The View parentView of this view (optionnal) */
    protected DialogView parentView;
    
    /** List of on change locale callbacks */
    private final List<Consumer<Locale>> intertionalizationCallbacks 
            = new ArrayList();
    
    /**
     * Creates new View form
     * @param parent
     * @throws java.lang.Exception
     */
    public DialogView(final Window parent) throws Exception {
        super(parent, null);
        setModal(true);
        this.initLookAndFeel();
        this.initContainer();
        this.initComponents();
        this.setLocale(AvailablesLocales.valueOf(
                ProjectProperties.getProperty("language",
                        AvailablesLocales.English.name())).getLocale());
    }

    public final void setParentView(final DialogView parent) {
        this.parentView = parent;
    }

    /**
     * Change the internationalization locale
     * @param locale
     */
    @Override
    public final void setLocale(final Locale locale) {
        try {
            /* Change the current locale */
            super.setLocale(locale);
                        
            /* Load the bundle associated with the new locale */
            this.bundle = java.util.ResourceBundle.getBundle
                    ("locales/interface", locale);
            
            /* Call each locale callbacks */
            if (intertionalizationCallbacks != null) {
                for (Consumer<Locale> callback : intertionalizationCallbacks) {
                    callback.accept(locale);
                }
            }
        } catch (Exception ex) {
            /* Ignore exceptions */
        }
    }
    
    /**
     * Add a callback to call when the locale is changed
     * @param callback 
     */
    protected void addLocaleCallback(final Consumer<Locale> callback) {
        intertionalizationCallbacks.add(callback);
    }
    
    /**
     * Load an icon
     * @param filePath
     * @return 
     */
    protected final ImageIcon loadIcon(final String filePath) {
        ImageIcon imageIcon = new ImageIcon();
        
        try {
            imageIcon = new javax.swing.ImageIcon(getClass().getResource(filePath));
        } catch (NullPointerException e) {
            
        }
        
        return imageIcon;
    }
    
    /**
     * Set the Nimbus look and feel
     */
    private void initLookAndFeel() throws Exception {
        javax.swing.UIManager.setLookAndFeel(
                    "javax.swing.plaf.metal.MetalLookAndFeel");
    }

    /**
     * Initialize the container options
     * @throws java.lang.Exception
     */
    protected abstract void initContainer() throws Exception;
    
    /**
     * Generate components
     * @throws java.lang.Exception
     */
    protected abstract void initComponents() throws Exception;
    
}
