/*
 * This software is governed by the CeCILL-B license under French law and
 * abiding by the rules of distribution of free software. You can use, modify
 * and/ or redistribute the software under the terms of the CeCILL-B license as
 * circulated by CEA, CNRS and INRIA at the following URL "http://www.cecill.info".
 * 
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability.
 * 
 * In this respect, the user's attention is drawn to the risks associated with
 * loading, using, modifying and/or developing or reproducing the software by
 * the user in light of its specific status of free software, that may mean that
 * it is complicated to manipulate, and that also therefore means that it is
 * reserved for developers and experienced professionals having in-depth
 * computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling
 * the security of their systems and/or data to be ensured and, more generally,
 * to use and operate it in the same conditions as regards security.
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.amap.lidar.amapvox.nephos3d.resources.meshes;

import com.jogamp.common.nio.Buffers;
import com.jogamp.opengl.GL3bc;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import fr.amap.lidar.amapvox.nephos3d.resources.Resource;
import net.jcip.annotations.GuardedBy;
import net.jcip.annotations.ThreadSafe;

/**
 * The mesh class is combined with a 
 * geometry to display it with rendering options.
 */
@ThreadSafe
public final class Mesh extends Resource {
        
    /**
     * The geometry object of this mesh.
     */
    @GuardedBy("this")
    private Geometry geometry;
    
    /**
     * Which shape to draw the primitive.
     */
    @GuardedBy("this")
    private DrawingType drawingType;
    
    /**
     * How many primitives to draw.
     */
    @GuardedBy("this")
    private int instances;
    
    /**
     * If the back faces are hiden or not (by default true).
     */
    @GuardedBy("this")
    private boolean faceCulling;
    
    /**
     * The depth function used with the depth buffer.
     */
    @GuardedBy("this")
    private DepthFunction depthFunction;
    
    /**
     * If the depth mask is used or not (by default true).
     */
    @GuardedBy("this")
    private boolean depthMask;
    
    /**
     * Vertex array object.
     */
    @GuardedBy("this")
    private final IntBuffer vao;
    
    /**
     * Vertex buffer object.
     */
    @GuardedBy("this")
    private final IntBuffer vbo;
    
    /**
     * Create a new mesh with en empty geometry, a drawing mode setted as 
     * "triangles", one instance, face culling and depth mask enabled and 
     * depth function setted as "less".
     * @param identifier 
     */
    public Mesh(final String identifier) {
        super(identifier);
        
        this.geometry = null;
        this.drawingType = DrawingType.TRIANGLES;
        this.instances = 1;
        this.faceCulling = true;
        this.depthMask = true;
        this.depthFunction = DepthFunction.LESS;
        
        this.vao = Buffers.newDirectIntBuffer(1);
        this.vbo = Buffers.newDirectIntBuffer(3);
    }

    /**
     * Chnage the current geometry.
     * @param geometry 
     */
    public final synchronized void setGeometry(final Geometry geometry) {
        this.geometry = geometry;
    }

    /**
     * Chnage the current drawing type.
     * @param drawingType 
     */
    public final synchronized void setDrawingType(final DrawingType drawingType) {
        this.drawingType = drawingType;
    }
    
    /**
     * Specify how many primitives to draw.
     * @param instances 
     */
    public final synchronized void setInstances(final int instances) {
        this.instances = instances;
    }

    /**
     * If the back faces are hiden or not (by default true).
     * @param faceCulling 
     */
    public final synchronized void setFaceCulling(final boolean faceCulling) {
        this.faceCulling = faceCulling;
    }

    /**
     * Set the depth function used with the depth buffer.
     * @param depthFunction 
     */
    public final synchronized void setDepthFunction(final DepthFunction depthFunction) {
        this.depthFunction = depthFunction;
    }

    /**
     * If the depth mask is used or not (by default true)
     * @param depthMask 
     */
    public final synchronized void setDepthMask(final boolean depthMask) {
        this.depthMask = depthMask;
    }
    
    @Override
    protected void onLoad(final GL3bc gl) {
        if (geometry != null) {
            /* Create vertex arrays */
            gl.glGenVertexArrays(vao.limit(), vao);
            gl.glBindVertexArray(vao.get(0));
            
            /* Create vertex buffers */
            gl.glGenBuffers(vbo.limit(), vbo);
            
            /* Create vertex position buffers */
            gl.glBindBuffer(GL3bc.GL_ARRAY_BUFFER, vbo.get(0));
            final FloatBuffer vertexBuffer = 
                    Buffers.newDirectFloatBuffer(geometry.getVertices());
            gl.glBufferData(GL3bc.GL_ARRAY_BUFFER,
                    vertexBuffer.limit() * Float.BYTES,
                    vertexBuffer, GL3bc.GL_STATIC_DRAW);
            
            /* Create vertex normal buffers */
            gl.glBindBuffer(GL3bc.GL_ARRAY_BUFFER, vbo.get(1));
            final FloatBuffer normalBuffer = 
                    Buffers.newDirectFloatBuffer(geometry.getNormals());
            gl.glBufferData(GL3bc.GL_ARRAY_BUFFER,
                    normalBuffer.limit() * Float.BYTES,
                    normalBuffer, GL3bc.GL_STATIC_DRAW);
            
            /* Create vertex normal buffers */
            gl.glBindBuffer(GL3bc.GL_ARRAY_BUFFER, vbo.get(2));
            final FloatBuffer texcoordBuffer = 
                    Buffers.newDirectFloatBuffer(geometry.getTexcoords());
            gl.glBufferData(GL3bc.GL_ARRAY_BUFFER,
                    texcoordBuffer.limit() * Float.BYTES,
                    texcoordBuffer, GL3bc.GL_STATIC_DRAW);
        }
    }
    
    @Override
    protected void onBind(final GL3bc gl) {
        if (geometry != null) {            
            /* Bind buffer and vertex array */
            gl.glBindVertexArray(vao.get(0));

            /* Enable vertex position attribute */
            gl.glBindBuffer(GL3bc.GL_ARRAY_BUFFER, vbo.get(0));
            gl.glEnableVertexAttribArray(0);
            gl.glVertexAttribPointer(0, 3, GL3bc.GL_FLOAT, false, 0, 0);

            /* Enable vertex normal attribute */
            gl.glBindBuffer(GL3bc.GL_ARRAY_BUFFER, vbo.get(1));
            gl.glEnableVertexAttribArray(1);
            gl.glVertexAttribPointer(1, 3, GL3bc.GL_FLOAT, false, 0, 0);

            /* Enable vertex texcoord attribute */
            gl.glBindBuffer(GL3bc.GL_ARRAY_BUFFER, vbo.get(2));
            gl.glEnableVertexAttribArray(2);
            gl.glVertexAttribPointer(2, 2, GL3bc.GL_FLOAT, false, 0, 0);
        }
    }
    
    /**
     * Draw this mesh by follow these steps :
     * <ol>
     *     <li>Bind the geometry.</li>
     *     <li>Update the face culling.</li>
     *     <li>Enable the depth test.</li>
     *     <li>Update the face depth mask.</li>
     *     <li>Draw each geometry instance with the current rendering method.</li>
     *     <li>Unbind the geometry.</li>
     * </ol>
     * @param gl The current OpenGL interface
     */
    public final synchronized void draw(final GL3bc gl) {
        if (geometry != null) {
            /* Bind the geometry */
            bind(gl);
            
            /* Hide back faces if face culling is enabled */
            if (faceCulling) {
                gl.glCullFace(GL3bc.GL_BACK);
                gl.glEnable(GL3bc.GL_CULL_FACE);
            } else {
                gl.glDisable(GL3bc.GL_CULL_FACE);
            }
            
            /* Change the depth function */
            switch(depthFunction) {
                case LEQUAL: 
                    gl.glDepthFunc(GL3bc.GL_LEQUAL);
                    break;
                case LESS:
                    gl.glDepthFunc(GL3bc.GL_LESS);
                    break;
            }
            
            /* Enabled blending */
            gl.glEnable(GL3bc.GL_BLEND);

            /* Change depth mask */
            gl.glDepthMask(depthMask);
            
            /* Check if the drawing mode is not null */
            if (null != drawingType) {
                /* Get the vertex indices */
                final IntBuffer indexBuffer = Buffers.newDirectIntBuffer(geometry.getIndices());
                
                /* Draw the available and selected drawing mode */
                switch (drawingType) {
                    case POINTS:
                        gl.glDrawArraysInstanced(GL3bc.GL_POINTS, 0, indexBuffer.limit(), instances);
                        break;
                    case LINES:
                        gl.glDrawArraysInstanced(GL3bc.GL_LINES, 0, indexBuffer.limit(), instances);
                        gl.glDrawArraysInstanced(GL3bc.GL_LINES_ADJACENCY, 0, indexBuffer.limit(), instances);
                        break;
                    case STRIP_LINES:
                        gl.glDrawArraysInstanced(GL3bc.GL_LINE_STRIP, 0, indexBuffer.limit(), instances);
                        break;
                    case TRIANGLES:
                        gl.glDrawArraysInstanced(GL3bc.GL_TRIANGLES, 0, indexBuffer.limit(), instances);
                        break;
                }
            }
            
            /* Unbind the geometry */
            unbind(gl);
        }
    }
    
    @Override
    protected void onUnbind(final GL3bc gl) {
        /* Unbind buffer and vertex array */
        if (geometry != null) {
            gl.glBindBuffer(GL3bc.GL_ARRAY_BUFFER, 0);
            gl.glBindVertexArray(0);
        }
    }

    @Override
    protected void onUnload(final GL3bc gl) {
        /* Clean up */
        if (geometry != null) {
            gl.glDeleteBuffers(vbo.capacity(), vbo);
            gl.glDeleteVertexArrays(vao.capacity(), vao);
        }
    }
    
}
